<?php

App::uses('EvShopAppModel', 'EvShop.Model');

class CategoryOverride extends EvShopAppModel {

	public $belongsTo = [
		'Category' => [
			'className' => 'EvShop.Category',
		],
		'Brand' => [
			'className' => 'EvShop.Brand',
		],
	];

	public $hasAndBelongsToMany = [
		'Option' => [
			'className' => 'EvShop.Option',
			'with' => 'EvShop.CategoryOverridesOption',
			'unique' => 'keepExisting',
		],
	];

	public $validate = [
		'category_id' => [
			'required' => [
				'rule' => 'notBlank',
				'message' => 'Select a category for the category override',
			],
		],
	];

/**
 * Format the options into HABTM data.
 *
 * @param array $data The data being saved.
 * @param array $options The save options.
 * @return array The modified data to be saved.
 */
	public function beforeBeforeSave($data, $options) {
		if (!empty($data['Option']) && is_array($data['Option'])) {
			$options = $data['Option'];

			//Clear out the data because if it isn't formatted we don't want to save it.
			unset($data['Option']);

			$selectedOptions = [];

			foreach ($options as $optionGroupId => $optionId) {
				//Select the optionId in the group
				if (!empty($optionId)) {
					$selectedOptions[] = $optionId;
				}
			}

			if (!empty($selectedOptions)) {
				$data['Option']['Option'] = $selectedOptions;
			} else {
				$data['Option']['Option'] = [];
			}
		}

		$overrideHash = $this->generateOverrideHash(
			!empty($data[$this->alias]['category_id']) ? $data[$this->alias]['category_id'] : null,
			!empty($data[$this->alias]['brand_id']) ? $data[$this->alias]['brand_id'] : null,
			!empty($data['Option']['Option']) ? $data['Option']['Option'] : []
		);

		if (!empty($overrideHash)) {
			$data[$this->alias]['override_hash'] = $overrideHash;
		}

		return $data;
	}

/**
 * Query used to retrieve a record ready for edit.
 *
 * @param int   $id ID of row to edit.
 * @param array $params The db query array - can be used to pass in additional parameters such as contain.
 * @return array.
 */
	public function readForEdit($id, $params = []) {
		$data = parent::readForEdit($id, $params);

		if (!empty($data[$this->alias]['id'])) {
			//Find and format the override options
			$options = $this->Option->find(
				'list',
				[
					'joins' => [
						[
							'table' => 'ev_shop_category_overrides_options',
							'alias' => 'CategoryOverridesOption',
							'conditions' => [
								'CategoryOverridesOption.option_id = ' . $this->Option->alias . '.id',
								'CategoryOverridesOption.category_override_id' => $id,
							],
						]
					],
					'fields' => [
						$this->Option->alias . '.option_group_id',
						$this->Option->alias . '.id',
					],
				]
			);

			$data[$this->Option->alias] = $options;
		}

		return $data;
	}

/**
 * Find an override based on the categoryId and a list of options. For an override to be found the list of options must
 * exactly match the list of options on the category override.
 *
 * @param int   $categoryId The id of the category to override.
 * @param int   $brandId    The id of the brand to match.
 * @param array $optionList A list of option ids to match.
 * @param array $params     Custom query parameters.
 * @return array The found category override.
 */
	public function findOverride($categoryId, $brandId, $optionList, $params = []) {
		$overrideHash = $this->generateOverrideHash($categoryId, $brandId, $optionList);

		if (empty($overrideHash)) {
			return [];
		}

		$defaultParams = [
			'conditions' => [
				$this->alias . '.override_hash' => $overrideHash,
			],
		];

		$params = Hash::merge($defaultParams, $params);

		return $this->find('first', $params);
	}

/**
 * Generate a hash for a category override using the category id, brand id and a list of options.
 *
 * @param int   $categoryId The id of the category the override is for.
 * @param int   $brandId    The id of the brand the override is for.
 * @param array $optionList A list of options the override is for.
 * @return null|string The generated hash, null if a hash could not be generated
 */
	public function generateOverrideHash($categoryId, $brandId, $optionList) {
		if (empty($categoryId)) {
			return null;
		}

		if (empty($categoryId) || !is_numeric($categoryId)) {
			$categoryId = '';
		}

		if (empty($brandId) || !is_numeric($brandId)) {
			$brandId = '';
		}

		if (empty($optionList) || !is_array($optionList)) {
			$optionList = [];
		}

		$hashString = $categoryId . ';' . $brandId . ';' . implode(',', $optionList);

		return md5($hashString);
	}
}
